import torch
import numpy as np
from transformers import T5Tokenizer, T5EncoderModel
import re
from tqdm import tqdm
import csv

class ProtBERT_embed:
    def __init__(self, device):
        self.device = device
        self.protbert_model = T5EncoderModel.from_pretrained("Rostlab/prot_t5_xl_half_uniref50-enc").to(device)
        self.tokenizer = T5Tokenizer.from_pretrained('Rostlab/prot_t5_xl_half_uniref50-enc', do_lower_case=False)
    
    def embed_seq(self, seq, batch_size, length=None):
        # use protBert to encode epitopeseq
        if length is None:
            max_length = len(max(seq, key=lambda x: len(x)))
        else:
            max_length = length
        seq_embed = torch.zeros(len(seq), max_length, 1024).float()
        epoch = np.ceil(len(seq)/batch_size).astype(int)
        for i in tqdm(range(epoch)):
            sequence_examples = seq[batch_size*i:batch_size*(i+1)]
            # replace all rare/ambiguous amino acids by X and introduce white-space between all amino acids
            sequence_examples = [" ".join(list(re.sub(r"[UZOB]", "X", sequence))) for sequence in sequence_examples]
            # tokenize sequences and pad up to the longest sequence in the batch
            ids = self.tokenizer.batch_encode_plus(sequence_examples, add_special_tokens=True, padding="longest")
            input_ids = torch.tensor(ids['input_ids']).to(self.device)
            attention_mask = torch.tensor(ids['attention_mask']).to(self.device)
            # generate embeddings
            with torch.no_grad():
                embedding_rpr = self.protbert_model(input_ids=input_ids,attention_mask=attention_mask)
            embed = embedding_rpr.last_hidden_state.cpu().detach()
            for j in range(len(embed)):
                seq_embed[batch_size*i+j,0:len(seq[batch_size*i+j]),:] = embed[j, 0:len(seq[batch_size*i+j])]
            torch.cuda.empty_cache()
        return seq_embed

class Atchley_embed:
    def __init__(self):
        pass

    def aamapping_TCR(self, peptideSeq,aa_dict,encode_dim):
        #Transform aa seqs to Atchley's factors.                                                                                              
        peptideArray = []
        if len(peptideSeq)>encode_dim:
            print('Length: '+str(len(peptideSeq))+' over bound!')
            peptideSeq=peptideSeq[0:encode_dim]
        for aa_single in peptideSeq:
            try:
                peptideArray.append(aa_dict[aa_single])
            except KeyError:
                print('Not proper aaSeqs: '+peptideSeq)
                peptideArray.append(np.zeros(5,dtype='float64'))
        for i in range(0,encode_dim-len(peptideSeq)):
            peptideArray.append(np.zeros(5,dtype='float64'))
        return np.asarray(peptideArray)

    def TCRMap(self, dataset,aa_dict,encode_dim):
        #Wrapper of aamapping                                                                                                                 
        for i in range(0,len(dataset)):
            if i==0:
                TCR_array=self.aamapping_TCR(dataset[i],aa_dict,encode_dim).reshape(1,encode_dim,5)
            else:
                TCR_array=np.append(TCR_array,self.aamapping_TCR(dataset[i],aa_dict,encode_dim).reshape(1,encode_dim,5),axis=0)
        print('TCRMap done!')
        return TCR_array
    
    def embed_seq(self, seq, encode_dim):
        aa_dict_dir='../Data/Atchley_factors.csv'
        aa_dict_atchley={}
        with open(aa_dict_dir,'r') as aa:
            aa_reader=csv.reader(aa)
            next(aa_reader, None)
            for rows in aa_reader:
                aa_name=rows[0]
                aa_factor=rows[1:len(rows)]
                aa_dict_atchley[aa_name]=np.asarray(aa_factor,dtype='float')
                
        seq_embed = self.TCRMap(seq, aa_dict_atchley, encode_dim)
        return torch.from_numpy(seq_embed).float()
        


import pandas as pd
import csv
import numpy as np
from io import StringIO
import torch
import re

def _file2dict(filename,key_fields,store_fields,delimiter='\t'):
    """Read file to a dictionary.
    key_fields: fields to be used as keys
    store_fields: fields to be saved as a list
    delimiter: delimiter used in the given file."""
    dictionary={}
    with open(filename, newline='') as csvfile:
        reader = csv.DictReader(csvfile,delimiter=delimiter)
        for row in reader:
            keys = [row[k] for k in key_fields]
            store= [row[s] for s in store_fields]

            sub_dict = dictionary
            for key in keys[:-1]:
                if key not in sub_dict:
                    sub_dict[key] = {}
                sub_dict = sub_dict[key]
            key = keys[-1]
            if key not in sub_dict:
                sub_dict[key] = []
            sub_dict[key].append(store)
    return dictionary

def _get_protseqs_ntseqs(chain='B', library_dir='library/'):
    """returns sequence dictioaries for genes: protseqsV, protseqsJ, nucseqsV, nucseqsJ"""
    seq_dicts=[]
    for gene,type in zip(['v','j','v','j'],['aa','aa','nt','nt']):
        name = library_dir+'tr'+chain.lower()+gene+'s_'+type+'.tsv'
        sdict = _file2dict(name,key_fields=['Allele'],store_fields=[type+'_seq'])
        for g in sdict:
            sdict[g]=sdict[g][0][0]
        seq_dicts.append(sdict)
    return seq_dicts

def determine_tcr_seq_vj(cdr3, V, J, chain, guess01=False, library_dir='library/'):
    if chain not in ['A', 'B']:
        raise ValueError("Chain must be 'A' or 'B'")

    protV, protJ = _get_protseqs_ntseqs(chain=chain, library_dir=library_dir)[:2]

    tcr_list = []

    for cdr3_, V_, J_ in zip(cdr3, V, J):
        if guess01:
            if '*' not in V_:
                V_ += '*01'
            if '*' not in J_:
                J_ += '*01'
        try:
            pv = protV[V_]
            pj = protJ[J_]
        except KeyError as e:
            raise KeyError(f"Cannot find {e.args[0]} in library for chain {chain}.")

        v_prefix = pv[:pv.rfind('C')]

        match = re.search(r'[FW]G.G', pj)
        if not match:
            raise ValueError(f"Pattern [FW]G.G not found in J protein {J_}.")
        j_suffix = pj[match.start() + 1:]

        tcr_seq = v_prefix + cdr3_ + j_suffix
        tcr_list.append(tcr_seq)

    return tcr_list

        
########################### One Hot ##########################   
aa_dict_one_hot = {'A': 0,'C': 1,'D': 2,'E': 3,'F': 4,'G': 5,'H': 6,'I': 7,'K': 8,'L': 9,
           'M': 10,'N': 11,'P': 12,'Q': 13,'R': 14,'S': 15,'T': 16,'V': 17,
           'W': 18,'Y': 19,'X': 20} 
########################### Blosum ########################## 
BLOSUM50_MATRIX = pd.read_table(StringIO(u"""                                                                                      
    A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V  B  J  Z  X  *                                                           
    A  5 -2 -1 -2 -1 -1 -1  0 -2 -1 -2 -1 -1 -3 -1  1  0 -3 -2  0 -2 -2 -1 -1 -5                                                           
    R -2  7 -1 -2 -4  1  0 -3  0 -4 -3  3 -2 -3 -3 -1 -1 -3 -1 -3 -1 -3  0 -1 -5                                                           
    N -1 -1  7  2 -2  0  0  0  1 -3 -4  0 -2 -4 -2  1  0 -4 -2 -3  5 -4  0 -1 -5                                                           
    D -2 -2  2  8 -4  0  2 -1 -1 -4 -4 -1 -4 -5 -1  0 -1 -5 -3 -4  6 -4  1 -1 -5                                                           
    C -1 -4 -2 -4 13 -3 -3 -3 -3 -2 -2 -3 -2 -2 -4 -1 -1 -5 -3 -1 -3 -2 -3 -1 -5                                                           
    Q -1  1  0  0 -3  7  2 -2  1 -3 -2  2  0 -4 -1  0 -1 -1 -1 -3  0 -3  4 -1 -5                                                           
    E -1  0  0  2 -3  2  6 -3  0 -4 -3  1 -2 -3 -1 -1 -1 -3 -2 -3  1 -3  5 -1 -5                                                           
    G  0 -3  0 -1 -3 -2 -3  8 -2 -4 -4 -2 -3 -4 -2  0 -2 -3 -3 -4 -1 -4 -2 -1 -5                                                           
    H -2  0  1 -1 -3  1  0 -2 10 -4 -3  0 -1 -1 -2 -1 -2 -3  2 -4  0 -3  0 -1 -5                                                          
    I -1 -4 -3 -4 -2 -3 -4 -4 -4  5  2 -3  2  0 -3 -3 -1 -3 -1  4 -4  4 -3 -1 -5                                                           
    L -2 -3 -4 -4 -2 -2 -3 -4 -3  2  5 -3  3  1 -4 -3 -1 -2 -1  1 -4  4 -3 -1 -5                                                           
    K -1  3  0 -1 -3  2  1 -2  0 -3 -3  6 -2 -4 -1  0 -1 -3 -2 -3  0 -3  1 -1 -5                                                           
    M -1 -2 -2 -4 -2  0 -2 -3 -1  2  3 -2  7  0 -3 -2 -1 -1  0  1 -3  2 -1 -1 -5                                                           
    F -3 -3 -4 -5 -2 -4 -3 -4 -1  0  1 -4  0  8 -4 -3 -2  1  4 -1 -4  1 -4 -1 -5                                                           
    P -1 -3 -2 -1 -4 -1 -1 -2 -2 -3 -4 -1 -3 -4 10 -1 -1 -4 -3 -3 -2 -3 -1 -1 -5                                                           
    S  1 -1  1  0 -1  0 -1  0 -1 -3 -3  0 -2 -3 -1  5  2 -4 -2 -2  0 -3  0 -1 -5                                                           
    T  0 -1  0 -1 -1 -1 -1 -2 -2 -1 -1 -1 -1 -2 -1  2  5 -3 -2  0  0 -1 -1 -1 -5                                                           
    W -3 -3 -4 -5 -5 -1 -3 -3 -3 -3 -2 -3 -1  1 -4 -4 -3 15  2 -3 -5 -2 -2 -1 -5                                                           
    Y -2 -1 -2 -3 -3 -1 -2 -3  2 -1 -1 -2  0  4 -3 -2 -2  2  8 -1 -3 -1 -2 -1 -5                                                           
    V  0 -3 -3 -4 -1 -3 -3 -4 -4  4  1 -3  1 -1 -3 -2  0 -3 -1  5 -3  2 -3 -1 -5                                                           
    B -2 -1  5  6 -3  0  1 -1  0 -4 -4  0 -3 -4 -2  0  0 -5 -3 -3  6 -4  1 -1 -5                                                           
    J -2 -3 -4 -4 -2 -3 -3 -4 -3  4  4 -3  2  1 -3 -3 -1 -2 -1  2 -4  4 -3 -1 -5                                                           
    Z -1  0  0  1 -3  4  5 -2  0 -3 -3  1 -1 -4 -1  0 -1 -2 -2 -3  1 -3  5 -1 -5                                                           
    X -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -5                                                           
    * -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5 -5  1                                                           
"""), sep='\s+').loc[list(aa_dict_one_hot.keys()), list(aa_dict_one_hot.keys())]
assert (BLOSUM50_MATRIX == BLOSUM50_MATRIX.T).all().all()

ENCODING_DATA_FRAMES = {
    "BLOSUM50": BLOSUM50_MATRIX
}

def _aa_encode(sequence, maxlen, encoding_method):
    sequence = sequence.replace(u'\xa0', u'').upper()
    if len(sequence) > maxlen:
        raise ValueError(f"Sequence {sequence} is longer than maxlen {maxlen}.")

    # convert amino acids to indices
    indices = [aa_dict_one_hot.get(aa, 20) for aa in sequence]  # 20 = 'X' padding
    indices += [20] * (maxlen - len(indices))  # padding
    result = ENCODING_DATA_FRAMES[encoding_method].iloc[indices]
    return np.asarray(result)

def amino_acid_encode(dataset, maxlen, encoding_method='BLOSUM50', encode_func=_aa_encode):
    pos = 0
    array = np.zeros((len(dataset), maxlen, 21), dtype=np.float32)
    cache = {}
    for item in dataset:
        if item not in cache:
            array[pos] = encode_func(item, maxlen, encoding_method).reshape(1, maxlen, 21)
            cache[item] = array[pos]
        else:
            array[pos] = cache[item]
        pos += 1
    return array

def process_phmc(df, pep_max_len=14, hla_max_len=34):

    epitope_array = amino_acid_encode(df['Epitope'].tolist(), pep_max_len)
    HLA_array = amino_acid_encode(df['HLA_aa'].tolist(), hla_max_len)
    
    epitope_tensor = torch.from_numpy(epitope_array).transpose(1,2).float()
    HLA_tensor = torch.from_numpy(HLA_array).transpose(1,2).float()
    
    return df, epitope_tensor, HLA_tensor

def negative_sampling(df, negative_ratio=3):
    original_tuples = set(map(tuple, df.to_numpy()))
    cols = ['alpha', 'beta', 'V_alpha', 'J_alpha', 'V_beta', 'J_beta']
    negative_samples = []

    for _ in range(negative_ratio):
        shuffled = df[cols].sample(frac=1).reset_index(drop=True)
        shuffled_df = df.copy()
        shuffled_df[cols] = shuffled.values

        new_samples = shuffled_df[~shuffled_df.apply(tuple, axis=1).isin(original_tuples)]
        negative_samples.append(new_samples)

    return pd.concat(negative_samples, ignore_index=True)



import torch
import torch.nn as nn
import numpy as np
import torch.optim as optim
from tqdm import tqdm
from collections import defaultdict
from torch.nn.utils.parametrizations import weight_norm
from tqdm import tqdm
from trimap import net
from trimap import utils
import pandas as pd
import logging
import os
from sklearn.metrics import roc_auc_score

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
    
class HLA_vae(nn.Module):
    """
    Variational Autoencoder (VAE) model for HLA sequences.

    This model learns a latent representation of HLA amino acid sequences
    and reconstructs the original sequence through a decoder. It is trained
    to minimize reconstruction loss and KL divergence.

    Args:
        input_size (list[int]): Shape of input tensor (typically [channels, length]).
        latent_size (int): Dimension of latent representation.
    """
    def __init__(self, input_size, latent_size):
        super(HLA_vae, self).__init__()
        
        self.HLA_encoder = net.Encoder(input_size=input_size,hidden=latent_size)
        self.HLA_Linear = net.Linear_block(latent_size, latent_size, latent_size)
        self.HLA_decoder = net.Decoder(hidden=latent_size,output_size=input_size)
        
    def train_model(
        self,
        train_loader,
        lr,
        epochs,
        device,
    ):
        """
        Train the HLA VAE model using MSE loss and KL divergence.

        Args:
            train_loader (DataLoader): Dataloader providing batches of HLA input tensors.
            lr (float): Learning rate.
            epochs (int): Number of training epochs.
            device (torch.device): Device to run training on (CPU or GPU).
        """
        self.train()
        
        optimizer = optim.Adam(self.parameters(), lr=lr, weight_decay=1e-4)
        MSELoss = nn.MSELoss()
        
        with tqdm(range(epochs), total=epochs, desc='Epochs') as tq:
            
            for epoch in tq:
                    
                epoch_loss = defaultdict(float)
                for i, hla in enumerate(train_loader):
                    
                    hla = hla[0].to(device)
                    
                    z_hla, _ = self.HLA_encoder(hla)
                    
                    z_hla, z_hla_mean, z_hla_logvar, x_hla = self.HLA_Linear(z_hla)
                    
                    hla_hat = self.HLA_decoder(x_hla)
                    
                    recon_loss = MSELoss(hla_hat, hla)
                    kl_loss = net.kl_div(z_hla_mean, z_hla_logvar)
                    
                    loss = {'KL_loss':1e-4*kl_loss, 'recon_loss':recon_loss}
                    
                    optimizer.zero_grad()
                    sum(loss.values()).backward()
                    
                    optimizer.step()
                    
                    for k,v in loss.items():
                        epoch_loss[k] += loss[k].item()
                        
                epoch_loss = {k:v/(i+1) for k, v in epoch_loss.items()}
                epoch_info = ','.join(['{}={:.4f}'.format(k, v) for k,v in epoch_loss.items()])
                tq.set_postfix_str(epoch_info)
        
    def embed_hla(
        self,
        hla_loader,
        device
    ):
        """
        Generate latent embeddings for HLA sequences using the trained encoder.

        Args:
            hla_loader (DataLoader): Dataloader of input HLA sequences.
            device (torch.device): Device for model computation.

        Returns:
            Tuple[np.ndarray, np.ndarray]: 
                - z_hla_mean: Latent mean vectors for HLA.
                - z_conv: Convolutional feature embeddings.
        """
        hla_embed = []
        conv_embed = []
        
        self.eval()

        with torch.no_grad():
            
            for i, data in enumerate(hla_loader):
                
                hla = data[0].to(device)
            
                z_hla, z_conv = self.HLA_encoder(hla)
                z_hla_mean, z_hla_logvar, x_hla = self.HLA_Linear(z_hla)
                
                hla_embed.append(z_hla_mean.detach().cpu().numpy())
                conv_embed.append(z_conv.detach().cpu().numpy())
                
        return np.concatenate(hla_embed, axis=0), np.concatenate(conv_embed, axis=0)
        
class PEP_vae(nn.Module):
    """
    Variational Autoencoder (VAE) for peptide (epitope) sequences.

    This model learns a latent encoding of peptide sequences and decodes them.
    It also supports alignment loss with HLA embeddings to enforce biological consistency.

    Args:
        input_size (list[int]): Shape of peptide input (e.g., [21, max_pep_length]).
        latent_size (int): Dimension of latent space.
    """
    def __init__(self, input_size, latent_size):
        super(PEP_vae, self).__init__()
        
        self.epitope_encoder = net.Encoder(input_size=input_size,hidden=latent_size)
        self.epitope_Linear = net.Linear_block(latent_size, latent_size, latent_size)
        self.epitope_decoder = net.Decoder(hidden=latent_size,output_size=input_size)
        
    def train_model(
        self,
        train_loader,
        lr,
        epochs,
        device,
    ):
        """
        Train the peptide VAE using reconstruction, KL, and alignment losses.

        Args:
            train_loader (DataLoader): Dataloader of (peptide, HLA) pairs.
            lr (float): Learning rate.
            epochs (int): Number of training epochs.
            device (torch.device): Target device for training.
        """
        self.train()
        
        optimizer = optim.Adam(self.parameters(), lr=lr, weight_decay=1e-4)
        MSELoss = nn.MSELoss()
        
        with tqdm(range(epochs), total=epochs, desc='Epochs') as tq:
            
            for epoch in tq:
                    
                epoch_loss = defaultdict(float)
                
                for i, data in enumerate(train_loader):
                    
                    pep, z_hla = data
                    pep, z_hla = pep.to(device), z_hla.to(device)
                    
                    z_pep, _ = self.epitope_encoder(pep)
                    
                    z_pep, z_pep_mean, z_pep_logvar, x_pep = self.epitope_Linear(z_pep)
                    
                    pep_hat = self.epitope_decoder(x_pep)
                    
                    recon_loss = MSELoss(pep_hat, pep)
                    kl_loss = net.kl_div(z_pep_mean, z_pep_logvar)
                    
                    align_loss = MSELoss(z_pep_mean, z_hla)
                    
                    loss = {'KL_loss':1e-4*kl_loss, 'align_loss':5*align_loss, 'recon_loss':recon_loss}

                    optimizer.zero_grad()
                    sum(loss.values()).backward()
                    
                    optimizer.step()
                    
                    for k,v in loss.items():
                        epoch_loss[k] += loss[k].item()
                        
                epoch_loss = {k:v/(i+1) for k, v in epoch_loss.items()}
                epoch_info = ','.join(['{}={:.4f}'.format(k, v) for k,v in epoch_loss.items()])
                tq.set_postfix_str(epoch_info) 
                    
    def embed_pep(
            self,
            data_loader,
            device
    ):
        """
        Embed peptides into latent space using the trained encoder.

        Args:
            data_loader (DataLoader): Input peptide batches.
            device (torch.device): Device for inference.

        Returns:
            Tuple[np.ndarray, np.ndarray]: 
                - z_pep_mean: Latent peptide embeddings.
                - z_conv: Intermediate convolutional features.
        """
        pep_embed = []
        conv_embed = []
        
        self.eval()

        with torch.no_grad():
            for i, data in enumerate(data_loader):
                
                data = data[0].to(device)
                
                z_pep, z_conv = self.epitope_encoder(data)
                
                z_pep_mean, _, _ = self.epitope_Linear(z_pep)
                
                pep_embed.append(z_pep_mean.detach().cpu().numpy())
                conv_embed.append(z_conv.detach().cpu().numpy())

        return np.concatenate(pep_embed, axis=0), np.concatenate(conv_embed, axis=0)

class TCRbind(nn.Module):
    """
    THE (TCR-HLA-Epitope) model for predicting T-cell epitope binding.

    This model integrates TCR alpha/beta chain embeddings with target embeddings (HLA or peptide),
    uses Bilinear Attention Network (BAN) layers, and outputs a classification score.

    Args:
        ban_heads (int): Number of attention heads in BAN.
        mlp (list[int]): MLP layer sizes [input, hidden, output].
        v_dim (int): Feature dimension of the target (e.g., peptide or HLA).
        q_dim (int): Feature dimension of TCR chains.
    """
    def __init__(self, ban_heads=8, mlp=[512, 128, 32], v_dim=256, q_dim=1024):
        super(TCRbind, self).__init__()
        self.bcn_alpha = weight_norm(net.BANLayer(v_dim=v_dim, q_dim=q_dim, h_dim=mlp[0], h_out=ban_heads), name='h_mat', dim=None)
        self.bcn_beta = weight_norm(net.BANLayer(v_dim=v_dim, q_dim=q_dim, h_dim=mlp[0], h_out=ban_heads), name='h_mat', dim=None)
        self.classifier = net.MLPDecoder(in_dim=mlp[0]*2, hidden_dim=mlp[1], out_dim=mlp[2], binary=1)
        self.alpha_dict = None
        self.beta_dict = None

    def print_logo(self):
        """Print ASCII logo banner to console."""
        logo = r"""
 ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄       ▄▄  ▄▄▄▄▄▄▄▄▄▄▄  ▄▄▄▄▄▄▄▄▄▄▄ 
▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌▐░░▌     ▐░░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌
 ▀▀▀▀█░█▀▀▀▀ ▐░█▀▀▀▀▀▀▀█░▌ ▀▀▀▀█░█▀▀▀▀ ▐░▌░▌   ▐░▐░▌▐░█▀▀▀▀▀▀▀█░▌▐░█▀▀▀▀▀▀▀█░▌
     ▐░▌     ▐░▌       ▐░▌     ▐░▌     ▐░▌▐░▌ ▐░▌▐░▌▐░▌       ▐░▌▐░▌       ▐░▌
     ▐░▌     ▐░█▄▄▄▄▄▄▄█░▌     ▐░▌     ▐░▌ ▐░▐░▌ ▐░▌▐░█▄▄▄▄▄▄▄█░▌▐░█▄▄▄▄▄▄▄█░▌
     ▐░▌     ▐░░░░░░░░░░░▌     ▐░▌     ▐░▌  ▐░▌  ▐░▌▐░░░░░░░░░░░▌▐░░░░░░░░░░░▌
     ▐░▌     ▐░█▀▀▀▀█░█▀▀      ▐░▌     ▐░▌   ▀   ▐░▌▐░█▀▀▀▀▀▀▀█░▌▐░█▀▀▀▀▀▀▀▀▀ 
     ▐░▌     ▐░▌     ▐░▌       ▐░▌     ▐░▌       ▐░▌▐░▌       ▐░▌▐░▌          
     ▐░▌     ▐░▌      ▐░▌  ▄▄▄▄█░█▄▄▄▄ ▐░▌       ▐░▌▐░▌       ▐░▌▐░▌          
     ▐░▌     ▐░▌       ▐░▌▐░░░░░░░░░░░▌▐░▌       ▐░▌▐░▌       ▐░▌▐░▌          
      ▀       ▀         ▀  ▀▀▀▀▀▀▀▀▀▀▀  ▀         ▀  ▀         ▀  ▀           
        """
        print(logo)

    def load_dict(self, device, df_data, max_alpha, max_beta, re_embed):
        """
        Load or create TCR α/β embedding dictionaries using ProtBERT.

        Args:
            device (torch.device): Target device for embedding.
            df_data (pd.DataFrame): DataFrame with 'alpha' and 'beta' sequences.
            max_alpha (int): Maximum alpha sequence length.
            max_beta (int): Maximum beta sequence length.
            re_embed (bool): If True, force re-embedding even if cache exists.
        """
        from trimap.BERT import ProtBERT_embed as PB
        prot_bert = PB(device)

        # Generalized loader/creator for a chain ('alpha' or 'beta')
        def get_or_create_dict(chain, max_len):
            path = f'{chain}_dict.pt'
            if os.path.exists(path) and not re_embed:
                logger.info(f'Loading {path}')
                chain_dict = torch.load(path)
            else:
                logger.info(f'{path} not found, generating...')
                unique_seqs = df_data[chain].unique()
                embeddings = prot_bert.embed_seq(unique_seqs, batch_size=256, length=max_len).cpu()
                chain_dict = dict(zip(unique_seqs, embeddings))
                torch.save(chain_dict, path)
                logger.info(f'Saved {path}')
            return chain_dict

        # Add new sequences not in existing dict
        def update_with_new_seqs(chain_dict, chain, max_len):
            all_seqs = df_data[chain].unique()
            new_seqs = [s for s in all_seqs if s not in chain_dict]
            if not new_seqs:
                logger.info(f'No new {chain} sequences found')
                return chain_dict

            logger.info(f'Found new {chain} sequences, embedding...')
            new_embeddings = prot_bert.embed_seq(new_seqs, batch_size=256, length=max_len).cpu()
            for seq, emb in zip(new_seqs, new_embeddings):
                chain_dict[seq] = emb
            torch.save(chain_dict, f'{chain}_dict.pt')
            logger.info(f'Updated and saved {chain}_dict.pt')
            return chain_dict

        # Process alpha and beta chains
        self.alpha_dict = update_with_new_seqs(get_or_create_dict('alpha', max_alpha), 'alpha', max_alpha)
        self.beta_dict = update_with_new_seqs(get_or_create_dict('beta', max_beta), 'beta', max_beta)
        
    def create_dataloader(
        self, 
        df_data,
        batch_size,
        device, 
        target, 
        mode, 
        max_alpha,
        max_beta,
        max_pep,
        pep_model_dir='pep_model.pt',
        hla_model_dir='hla_model.pt',
        hla_dict=None,
        load_dict=True,
        re_embed=False
    ):
        """
        Create DataLoader for training or testing the THE model.

        Depending on `target`, the function embeds HLA or peptide inputs,
        and returns batched tensors including TCR and target embeddings.

        Args:
            df_data (pd.DataFrame): Data with columns for alpha, beta, epitope/HLA, etc.
            batch_size (int): Batch size for DataLoader.
            device (torch.device): Computation device.
            target (str): 'peptide' or 'hla'.
            mode (str): 'train' or 'test'.
            max_alpha (int): Maximum length for alpha CDR3 sequences.
            max_beta (int): Maximum length for beta CDR3 sequences.
            max_pep (int): Maximum length for peptide sequences (if target is peptide).
            pep_model_dir (str): Path to pre-trained peptide VAE model.
            hla_model_dir (str): Path to pre-trained HLA VAE model.
            hla_dict (dict, optional): Dictionary mapping HLA names to amino acid sequences.
            load_dict (bool): If True, load/create TCR α/β embedding dictionaries.
            re_embed (bool): If True, force re-embedding of TCR sequences.
        Raises:
            ValueError: If `target` is not 'peptide' or 'hla'.
            ValueError: If `mode` is not 'train' or 'test'.
        
        Returns:
            torch.utils.data.DataLoader: Batched model input.
        """
        
        if load_dict:
            self.load_dict(device, df_data, max_alpha, max_beta, re_embed)
        
        alpha = df_data['alpha'].values.tolist()
        beta = df_data['beta'].values.tolist()
        alpha = [self.alpha_dict[i] for i in alpha]
        beta = [self.beta_dict[i] for i in beta]
        
        alpha = torch.stack(alpha)
        beta = torch.stack(beta)

        if target == 'peptide':
            epi_train = utils.amino_acid_encode(df_data['Epitope'].tolist(),max_pep)
            epi_train = torch.from_numpy(epi_train).transpose(1,2).float()
            
            pep_model = PEP_vae(input_size=[21, max_pep],latent_size=256).to(device)
            pep_model.load_state_dict(torch.load(pep_model_dir))
            
            dataset = torch.utils.data.TensorDataset(epi_train)
            dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=False)
            _, pep_embed = pep_model.embed_pep(dataloader, device=device)
            target_embed = torch.from_numpy(pep_embed).transpose(1, 2).float()
            
        elif target == 'hla':
            hla_aa = list(hla_dict.values())
            hla_name = list(hla_dict.keys())
            hla_name = ['HLA-'+i for i in hla_name]
            HLA_lib_array = utils.amino_acid_encode(hla_aa, len(hla_aa[0]))
            HLA_lib_tensor = torch.from_numpy(HLA_lib_array).transpose(1,2).float()
            HLA_dict = {}
            for i in range(len(hla_name)):
                HLA_dict[hla_name[i]] = HLA_lib_tensor[i]
                
            HLA_embed = [HLA_dict[i] for i in df_data['HLA'].tolist()]
            HLA_embed = torch.stack(HLA_embed).float()
            
            hla_dataset = torch.utils.data.TensorDataset(HLA_embed)
            hla_test_loader = torch.utils.data.DataLoader(hla_dataset, batch_size=batch_size, shuffle=False)
            hla_model = HLA_vae(input_size=[21, len(hla_aa[0])], latent_size=256).to(device)
            hla_model.load_state_dict(torch.load(hla_model_dir))
            _, hla_embed = hla_model.embed_hla(hla_test_loader, device)
            target_embed = torch.from_numpy(hla_embed).transpose(1, 2).float()
        
        else:
            raise ValueError('target must be peptide or hla!')
        
        if mode == 'train':
            labels = torch.from_numpy(df_data['label'].values)
            dataset = torch.utils.data.TensorDataset(alpha, beta, target_embed, labels)
            shuffle = True
        elif mode == 'test':
            dataset = torch.utils.data.TensorDataset(alpha, beta, target_embed)
            shuffle = False
        else:
            raise ValueError("mode must be 'train' or 'test'")

        loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=8)
        return loader
    
    def _sampling(self, df_data):
        negative_samples = utils.negative_sampling(df_data)
        df_data['label'] = 1
        negative_samples['label'] = 0
        df_data = pd.concat([df_data, negative_samples])
        df_data.reset_index(drop=True, inplace=True)
        return df_data
    
    def train_model(
        self,
        df,
        num_epochs,
        device,
        batch_size=256,
        lr=1e-4,
        chain='ab',
        targets='peptide',
        hla_dict=None,
        df_add=None,
        neg_sampling=True,
        neg_resample=True,
        max_alpha=125,
        max_beta=127, 
        max_pep=14
    ):
        """
        Train the THE model to predict TCR-target binding.

        This function performs supervised training using TCR alpha/beta embeddings 
        and target embeddings (peptide or HLA). It optionally supports negative 
        sampling and dynamic resampling across epochs.

        Args:
            df (pd.DataFrame): Primary training dataset containing TCR sequences, 
                target peptides or HLAs, and binary labels.
            num_epochs (int): Number of training epochs.
            device (torch.device): Device to run training on (e.g., 'cuda:0' or 'cpu').
            batch_size (int): Batch size used in training. Default is 256.
            lr (float): Learning rate for the optimizer. Default is 1e-4.
            chain (str): Specifies which TCR chains to use in the model.
                Options are 'ab' (both), 'a', or 'b'. Default is 'ab'.
            targets (str): Specifies the type of target to embed — either 'peptide' or 'hla'.
            hla_dict (dict, optional): Mapping of HLA allele names to amino acid sequences.
                Required if targets='hla'.
            df_add (pd.DataFrame, optional): Additional training samples to be appended 
                (e.g., auxiliary positive data). Default is None.
            neg_sampling (bool): Whether to use negative sampling during training. Default is True.
            neg_resample (bool): If True, regenerates negative samples at each epoch. Default is True.
            max_alpha (int): Maximum length of alpha CDR3 sequences (for padding/embedding).
            max_beta (int): Maximum length of beta CDR3 sequences.
            max_pep (int): Maximum length of peptide sequences (if using peptide as target).

        Behavior:
            - If `neg_sampling` is True, negative samples are generated on-the-fly per epoch.
            - If `df_add` is provided, it is concatenated to the training set each epoch.
            - Model is trained using binary cross-entropy loss (`BCEWithLogitsLoss`).
            - ROC-AUC is computed each epoch for performance tracking.

        Returns:
            None. (Trained weights are updated in-place.)
        """
        self.print_logo()
        self.train()

        optimizer = torch.optim.Adam(self.parameters(), lr=lr, weight_decay=1e-4)
        loss_fct = nn.BCEWithLogitsLoss()
            
        df_data = df.copy()
        # If no negative sampling, prepare loader only once
        if not neg_sampling:
            if df_add is not None:
                df_data = pd.concat([df_data, df_add]).reset_index(drop=True)
                
            train_loader = self.create_dataloader(
                df_data, batch_size, device, targets, 'train',
                max_alpha=max_alpha, max_beta=max_beta, max_pep=max_pep, hla_dict=hla_dict,
            )

        logger.info('Training...')
        
        for epoch in range(num_epochs):
            # Prepare train loader if using negative sampling
            if neg_sampling and (epoch == 0 or neg_resample):
                df_data = self._sampling(df.copy())
                if df_add is not None:
                    df_data = pd.concat([df_data, df_add]).reset_index(drop=True)
                train_loader = self.create_dataloader(
                    df_data, batch_size, device, targets, 'train',
                    max_alpha=max_alpha, max_beta=max_beta, max_pep=max_pep, hla_dict=hla_dict,
                    load_dict=(epoch == 0)
                )

            scores_all, labels_all = [], []

            for data in tqdm(train_loader):
                if len(data[0]) == 1:
                    continue  # skip batch of size 1

                alpha, beta, target, labels = [x.float().to(device) for x in data]

                # Forward
                logits = self.forward_chain(target, alpha, beta, chain=chain, return_attn=False)
                scores = self.classifier(logits).squeeze()

                loss = loss_fct(scores, labels)

                # Backward
                optimizer.zero_grad()
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.parameters(), max_norm=5.0)
                optimizer.step()

                scores_all.append(scores.detach().cpu())
                labels_all.append(labels.cpu())

            # Evaluation
            scores = torch.cat(scores_all)
            labels = torch.cat(labels_all)
            roc = roc_auc_score(labels, scores)

            print(f'Epoch [{epoch+1}/{num_epochs}], Loss: {loss.item():.4f}, ROC: {roc:.4f}')
            torch.cuda.empty_cache()

    def forward_chain(self, target, alpha, beta, chain='ab', return_attn=False):
        """
        Perform forward computation for alpha/b with BAN attention.

        Args:
            target (Tensor): target embedding (e.g., peptide or HLA)
            alpha (Tensor): alpha chain embedding
            beta (Tensor): beta chain embedding
            chain (str): one of 'ab', 'a', or 'b'
            return_attn (bool): whether to return attention maps

        Returns:
            logits (Tensor): concatenated logits
            (Optional) attn_a, attn_b: attention weights if return_attn=True
        """
        attn_a = attn_b = None

        if chain == 'ab':
            a_logits, attn_a = self.bcn_alpha(target, alpha, softmax=True)
            b_logits, attn_b = self.bcn_beta(target, beta, softmax=True)
        elif chain == 'a':
            a_logits, attn_a = self.bcn_alpha(target, alpha, softmax=True)
            b_logits = torch.zeros_like(a_logits)
        elif chain == 'b':
            b_logits, attn_b = self.bcn_beta(target, beta, softmax=True)
            a_logits = torch.zeros_like(b_logits)
        else:
            raise ValueError("chain must be 'ab', 'a', or 'b'")

        logits = torch.cat([a_logits, b_logits], dim=1)

        if return_attn:
            return logits, attn_a, attn_b
        else:
            return logits
        
    def test_model(
        self,
        df_test=None,
        batch_size=256,
        test_loader=None,
        hla_dict=None,
        device='cuda:0',
        chain='ab',
        targets='peptide',
        max_alpha=125,
        max_beta=127,
        max_pep=14
    ):
        """
        Evaluate the THE model on a held-out test dataset.

        This function performs inference using the trained model and returns predicted 
        binding scores as well as attention weights (if applicable). Input can be provided 
        either as a DataFrame (`df_test`) or a pre-built `test_loader`.

        Args:
            df_test (pd.DataFrame, optional): Test dataset containing columns for 'alpha', 
                'beta', 'Epitope' or 'HLA'. Required if `test_loader` is not provided.
            batch_size (int): Batch size used during inference. Default is 256.
            test_loader (DataLoader, optional): Precomputed DataLoader for testing.
                If provided, `df_test` is ignored.
            hla_dict (dict, optional): Dictionary mapping HLA names to amino acid sequences.
                Required if `targets='hla'`.
            device (str or torch.device): Computation device (e.g., 'cuda:0' or 'cpu').
            chain (str): Specifies which TCR chain(s) to use: 'ab', 'a', or 'b'.
            targets (str): Specifies target type: 'peptide' or 'hla'.
            max_alpha (int): Maximum sequence length for alpha chains (for padding).
            max_beta (int): Maximum sequence length for beta chains.
            max_pep (int): Maximum peptide length (for encoding peptide targets).

        Returns:
            Tuple:
                - scores (torch.Tensor): Predicted probabilities (after sigmoid).
                - alpha_attn (torch.Tensor or None): Attention weights for alpha chain (if used).
                - beta_attn (torch.Tensor or None): Attention weights for beta chain (if used).
        """
        self.eval()

        if test_loader is None:
            test_loader = self.create_dataloader(
                df_test,
                batch_size=batch_size,
                device=device,
                target=targets,
                mode='test',
                max_alpha=max_alpha,
                max_beta=max_beta,
                max_pep=max_pep,
                 hla_dict=hla_dict,
            )

        logger.info('Predicting...')
        
        score_all = []
        attn_dict = {'a': [], 'b': []}

        with torch.no_grad():
            for data in tqdm(test_loader):
                alpha, beta, target = [x.float().to(device) for x in data[:3]]

                # Forward
                logits, attn_a, attn_b = self.forward_chain(target, alpha, beta, chain=chain, return_attn=True)
                scores = torch.sigmoid(self.classifier(logits).squeeze())

                score_all.append(scores.cpu())
                if attn_a is not None:
                    attn_dict['a'].append(attn_a.cpu())
                if attn_b is not None:
                    attn_dict['b'].append(attn_b.cpu())

        scores = torch.cat(score_all)

        alpha_attn = torch.cat(attn_dict['a'], dim=0) if attn_dict['a'] else None
        beta_attn = torch.cat(attn_dict['b'], dim=0) if attn_dict['b'] else None

        return scores, alpha_attn, beta_attn
    


import torch
import torch.nn as nn
from torch.nn.utils.parametrizations import weight_norm
    
class BANLayer(nn.Module):
    def __init__(self, v_dim, q_dim, h_dim, h_out, act='ReLU', dropout=0.2, k=3):
        super(BANLayer, self).__init__()

        self.c = 32
        self.k = k
        self.v_dim = v_dim
        self.q_dim = q_dim
        self.h_dim = h_dim
        self.h_out = h_out

        self.v_net = FCNet([v_dim, h_dim * self.k], act=act, dropout=dropout)
        self.q_net = FCNet([q_dim, h_dim * self.k], act=act, dropout=dropout)
        # self.dropout = nn.Dropout(dropout[1])
        if 1 < k:
            self.p_net = nn.AvgPool1d(self.k, stride=self.k)

        if h_out <= self.c:
            self.h_mat = nn.Parameter(torch.Tensor(1, h_out, 1, h_dim * self.k).normal_())
            self.h_bias = nn.Parameter(torch.Tensor(1, h_out, 1, 1).normal_())
        else:
            self.h_net = weight_norm(nn.Linear(h_dim * self.k, h_out), dim=None)

        self.bn = nn.BatchNorm1d(h_dim)

    def attention_pooling(self, v, q, att_map):
        fusion_logits = torch.einsum('bvk,bvq,bqk->bk', (v, att_map, q))
        if 1 < self.k:
            fusion_logits = fusion_logits.unsqueeze(1)  # b x 1 x d
            fusion_logits = self.p_net(fusion_logits).squeeze(1) * self.k  # sum-pooling
        return fusion_logits

    def forward(self, v, q, softmax=False, v_mask=True, q_mask=True, attn=None):
        v_num = v.size(1)
        q_num = q.size(1)
        if self.h_out <= self.c:
            v_ = self.v_net(v)
            q_ = self.q_net(q)
            att_maps = torch.einsum('xhyk,bvk,bqk->bhvq', (self.h_mat, v_, q_)) + self.h_bias
        else:
            v_ = self.v_net(v).transpose(1, 2).unsqueeze(3)
            q_ = self.q_net(q).transpose(1, 2).unsqueeze(2)
            d_ = torch.matmul(v_, q_)  # b x h_dim x v x q
            att_maps = self.h_net(d_.transpose(1, 2).transpose(2, 3))  # b x v x q x h_out
            att_maps = att_maps.transpose(2, 3).transpose(1, 2)  # b x h_out x v x q
        if v_mask:
            mask = (0 == v.abs().sum(2)).unsqueeze(1).unsqueeze(3).expand(att_maps.size())
            att_maps.data.masked_fill_(mask.data, -float('inf'))
        if q_mask:
            mask = (0 == q.abs().sum(2)).unsqueeze(1).unsqueeze(2).expand(att_maps.size())
            att_maps.data.masked_fill_(mask.data, -float('inf'))
        if softmax:
            p = nn.functional.softmax(att_maps.reshape(-1, self.h_out, v_num * q_num), 2)
            att_maps = p.view(-1, self.h_out, v_num, q_num)
        
        logits = self.attention_pooling(v_, q_, att_maps[:, 0, :, :])
        for i in range(1, self.h_out):
            logits_i = self.attention_pooling(v_, q_, att_maps[:, i, :, :])
            logits += logits_i
        logits = self.bn(logits)
        return logits, att_maps

    
class FCNet(nn.Module):
    def __init__(self, dims, act='ReLU', dropout=0.2):
        super(FCNet, self).__init__()

        layers = []
        for i in range(len(dims) - 2):
            in_dim = dims[i]
            out_dim = dims[i + 1]
            if 0 < dropout:
                layers.append(nn.Dropout(dropout))
            layers.append(weight_norm(nn.Linear(in_dim, out_dim), dim=None))
            if '' != act:
                layers.append(getattr(nn, act)())
        if 0 < dropout:
            layers.append(nn.Dropout(dropout))
        layers.append(weight_norm(nn.Linear(dims[-2], dims[-1]), dim=None))
        if '' != act:
            layers.append(getattr(nn, act)())

        self.main = nn.Sequential(*layers)

    def forward(self, x):
        return self.main(x)


class MLPDecoder(nn.Module):
    '''
    MLP decoder for TCR chain
    '''
    def __init__(self, in_dim, hidden_dim, out_dim, binary=1):
        super(MLPDecoder, self).__init__()
        self.fc1 = nn.Linear(in_dim, hidden_dim)
        self.bn1 = nn.BatchNorm1d(hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.bn2 = nn.BatchNorm1d(hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, out_dim)
        self.bn3 = nn.BatchNorm1d(out_dim)
        self.fc4 = nn.Linear(out_dim, binary)
        self.sigmoid = nn.Sigmoid()
        self.dropout = nn.Dropout(0)
        self.leakyrelu = nn.LeakyReLU()

    def forward(self, x):
        x = self.leakyrelu(self.bn1(self.fc1(x)))
        x = self.leakyrelu(self.bn2(self.fc2(x)))
        x = self.leakyrelu(self.bn3(self.fc3(x)))
        x = self.fc4(x)
        return x

def kl_div(mu, log_var):
    """
    KL divergence between the prior and posterior distributions
    """
    return torch.mean(-0.5 * torch.sum(1 + log_var - mu ** 2 - log_var.exp(), dim = 1), dim = 0)
    
class classifier(nn.Module):
    """
    Classifier for TCR chain
    """
    def __init__(self, input_size):
        super(classifier, self).__init__()
        
        self.seq = nn.Sequential(
            nn.Linear(input_size, 1024),
            nn.BatchNorm1d(1024),
            nn.LeakyReLU(),
            nn.Dropout(0.2),
            nn.Linear(1024, 256),
            nn.BatchNorm1d(256),
            nn.LeakyReLU(),
            nn.Dropout(0.2),
            nn.Linear(256, 64),
            nn.BatchNorm1d(64),
            nn.LeakyReLU(),
            nn.Dropout(0.2),
            nn.Linear(64, 1),
        )
        
    def forward(self, z):
        x = self.seq(z)
        return x

class Conv_block(torch.nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding, pool='max', drop=0.2):
        super(Conv_block, self).__init__()
    
        self.conv = nn.Conv1d(in_channels, out_channels, kernel_size, padding=padding)
        self.bn = nn.BatchNorm1d(out_channels)
        self.relu = nn.LeakyReLU()
        self.dropout = nn.Dropout(drop)
        
    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        x = self.dropout(x)
        return x
    
class Deconv_block(torch.nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding, pool='max', drop=0.2):
        super(Deconv_block, self).__init__()
    
        self.deconv = nn.ConvTranspose1d(in_channels, out_channels, kernel_size, padding=padding)
        self.bn = nn.BatchNorm1d(out_channels)
        self.relu = nn.LeakyReLU()
        self.dropout = nn.Dropout(drop)
        
    def forward(self, x):
        x = self.deconv(x)
        x = self.bn(x)
        x = self.relu(x)
        x = self.dropout(x)
        return x
    
class Encoder(nn.Module):
    """Convolution for TCR chain. """
    def __init__(self,input_size,hidden,filters=[64,128,256],kernel_sizes=[3,3,3],drop=0.2):
        super(Encoder, self).__init__()

        input_filters, length = input_size[0], input_size[1]
        
        self.cnn1 = Conv_block(input_filters, filters[0], kernel_size=kernel_sizes[0], padding=(kernel_sizes[0]-1)//2, drop=drop)
        self.cnn2 = Conv_block(filters[0], filters[1], kernel_size=kernel_sizes[1], padding=(kernel_sizes[1]-1)//2, drop=drop)
        self.cnn3 = Conv_block(filters[1], filters[2], kernel_size=kernel_sizes[2], padding=(kernel_sizes[2]-1)//2, drop=drop)
            
        self.seq = nn.Sequential(
            nn.Linear(filters[-1]*length, 1024),
            nn.BatchNorm1d(1024),
            nn.LeakyReLU(),
            nn.Dropout(drop),
            nn.Linear(1024, 512),
            nn.BatchNorm1d(512),
            nn.LeakyReLU(),
            nn.Dropout(drop),
            nn.Linear(512, hidden),
        )
            
    def forward(self,x):
        # CNN part
        x = self.cnn1(x)
        x = self.cnn2(x)
        xx = self.cnn3(x)
        x = xx.flatten(start_dim=1)
        x = self.seq(x)
        return x, xx
    
class Decoder(nn.Module):
    def __init__(self,hidden,output_size,filters=[256,128,64],kernel_sizes=[3,3,3,3],drop=0.2):
        super(Decoder, self).__init__()
        
        output_filters, length = output_size[0], output_size[1]
        
        self.t_linear = nn.Linear(hidden, 32*length) 
        self.deconv1 = Deconv_block(32, filters[0], kernel_size=kernel_sizes[0], padding=1, drop=drop)
        self.deconv2 = Deconv_block(filters[0], filters[1], kernel_size=kernel_sizes[1], padding=1, drop=drop)
        self.deconv3 = Deconv_block(filters[1], filters[2], kernel_size=kernel_sizes[2], padding=1, drop=drop)
        self.deconv4 = nn.ConvTranspose1d(filters[2], output_filters, kernel_size=kernel_sizes[3], padding=1)
        
    def forward(self,x):
        x = self.t_linear(x)
        x = x.view(x.shape[0],32,x.shape[1]//32)
        x = self.deconv1(x)
        x = self.deconv2(x)
        x = self.deconv3(x)
        x = self.deconv4(x)
        return x
    
class Linear_block(torch.nn.Module):
    def __init__(self, in_channels, out_channels, hidden, drop=0.2):
        super(Linear_block, self).__init__()
    
        self.linear1 = nn.Linear(in_channels, hidden)
        self.bn1 = nn.BatchNorm1d(hidden)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(drop)
        self.mean = nn.Linear(hidden,hidden)
        self.logvar = nn.Linear(hidden,hidden)
        self.linear2 = nn.Linear(hidden,out_channels)
        self.bn2 = nn.BatchNorm1d(out_channels)
        
    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5*logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add_(mu)
    
    def forward(self, x):
        x = self.dropout(self.relu(self.bn1(self.linear1(x))))
        z_mean = self.mean(x)
        z_logvar = self.logvar(x)
        if self.training:
            z = self.reparameterize(z_mean, z_logvar)
            x = self.dropout(self.relu(self.bn2(self.linear2(z))))
            return z, z_mean, z_logvar, x
        else:
            x = self.relu(self.bn2(self.linear2(z_mean)))
            return z_mean, z_logvar, x


from setuptools import setup, find_packages

setup(
    name="trimap-tools",
    version="1.0.1",
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=[
        "numpy",
        "pandas",
        "torch",
        "scikit-learn",
        "tqdm",
        "transformers",
        "matplotlib",
    ],
    author="Kai Cao",
    author_email="caokai1073@gmail.com",
    description="Trimap for predicting TCR and Peptide-HLA binding",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/caokai1073/themap-tools", 
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)